"""
Implementation of simulator for FederatedAveraging as described in the paper
Communication-Efficient Learning of Deep Networks from Decentralized Data
(McMahan, Brendan, et al. 2017)
http://proceedings.mlr.press/v54/mcmahan17a/mcmahan17a.pdf
 所以为了提高效率，我们可以把样本分成等量的子集。 例如我们把100万样本分成1000份， 每份1000个样本， 这些子集就称为mini batch。
 然后我们分别用一个for循环遍历这1000个子集。 针对每一个子集做一次梯度下降。 然后更新参数w和b的值。接着到下一个子集中继续进行梯度下降。
 这样在遍历完所有的mini batch之后我们相当于在梯度下降中做了1000次迭代。 我们将遍历一次所有样本的行为叫做一个 epoch，也就是一个世代。
 在mini batch下的梯度下降中做的事情其实跟full batch一样，只不过我们训练的数据不再是所有的样本，而是一个个的子集。
 这样在mini batch我们在一个epoch中就能进行1000次的梯度下降，而在full batch中只有一次。 这样就大大的提高了我们算法的运行速度。
"""

import numpy as np
import pandas as pd

import utils




class fedavg_worker(object):
    """
    Class to represent a FederatedAveraging worker (called a "Client" in the paper)
    """

    def __init__(self, model):
        """
        Initialize the worker with a Keras model

        Parameters
        ----------
        model : tf.keras.Model
            The model to be used by the worker to do local training.
        """

        # self.X_train_fridge = None
        # self.X_train_television = None
        # self.X_train_kettle = None
        # self.X_train_microwave = None
        # self.X_train_washer_dryer = None
        # self.X_train_HTPC = None
        # self.X_train_radio = None
        # self.X_train_computer = None
        # self.X_train_fan = None
        #
        # self.X_test_fridge = None
        # self.X_test_television = None
        # self.X_test_kettle = None
        # self.X_test_microwave = None
        # self.X_test_washer_dryer = None
        # self.X_test_HTPC = None
        # self.X_test_radio = None
        # self.X_test_computer = None
        # self.X_test_fan = None

        self.model = model
        self.X_train = None
        self.y_train = None
        self.X_test = None
        self.y_test = None
        # self.y_train_fridge = None
        # self.y_train_television = None
        # self.y_train_kettle = None
        # self.y_train_microwave = None
        # self.y_train_washer_dryer = None
        # self.y_train_HTPC = None
        # self.y_train_radio = None
        # self.y_train_computer = None
        # self.y_train_fan = None



class fedavg_server(object):
    """
    Class to represent a FederatedAveraging parameter server
    """

    def __init__(self, model, K):
        """
        Initialize the parameter server with a Keras model
        and a list of size K to store K worker weights

        Parameters
        ----------
        model : tf.keras.Model
            The model to be used by the parameter server as the global model.
        K : int
            The number of workers that can participate in federated learning.
        """
        self.global_model = model
        self.worker_weights = [None] * K


class fedavg_hparams(object):
    """
    Hyperparameters for the FederatedAveraging algorithm.
    """

    def __init__(self, K=5, C=1, E=1, B=100, eta=0.1, MAX_T=10000, iid=True,
                 evaluation_interval=5, target_val_accuracy=0.7):
        """
        Parameters
        ---------
        -----------------------------------------------------------
        The following variables are named to match the paper:

        K : int, optional
            The number of workers. The default is 100.
        C : float, optional
            The fraction of workers randomly selected per iteration. The default is 0.1.
        E : int, optional
            The number of epochs a worker will run over its data per iteration. The default is 1.
        B : int, optional
            The worker minibatch size. The default is 10.
        eta : float, optional
            Step size to use for worker SGD optimizers. The default is 0.1.
        -----------------------------------------------------------
        -----------------------------------------------------------
        The following variables are not assigned explicit names in the paper:

        MAX_T : int, optional
            Maximum number of rounds of FederatedAveraging to run. The default is 10000.
        iid : boolean, optional
            If iid, we simulate an evenly distributed random split of the data across
            workers. Otherwise each worker gets data in only one (or few) classes.
            The default is True.
        evaluation_interval : int, optional
            Evaluate global model whenever this many iterations of FederatedAveraging have been run. The default is 5.
        target_val_accuracy : float, optional
            Stop training if this target validation accuracy has been achieved. The default is 0.9.
        -----------------------------------------------------------
        """
        self.K = K
        self.C = C
        self.E = E
        self.B = B
        self.eta = eta
        self.MAX_T = MAX_T
        self.iid = iid
        self.evaluation_interval = evaluation_interval
        self.target_val_accuracy = target_val_accuracy


def federated_averaging(X_train, y_train, X_test, y_test, model_constructor, hparams, rng=None):
    """
    Simulate training a model using FederatedAveraging across K distributed devices.
    Return the final global model and metrics gathered over the course of the run.

    Parameters
    ----------
    X_train : numpy ndarray
        Training features.
    y_train : numpy ndarray
        Training targets.
    X_val : numpy ndarray
        Validation features.
    y_val : numpy ndarray
        Validation targets.
    model_constructor : function
        function that constructs a compiled tf.keras.Model using hparams.
    hparams : fedavg_hparams
        Hyperparameters for FederatedAveraging.
    rng : numpy.random.Generator, optional
       instance to use for random number generation.

    Returns
    -------
    global_model : tf.keras.Model
        The final global model

    log : dict
        Dictionary containing training and validation metrics:
            loss :
                training loss at each iteration
            accuracy :
                training accuracy at each iteration
            val_loss :
                validation loss at each iteration
            val_accuracy :
                validation accuracy at each iteration
            iteration :
                the iteration number at which the measurements were made
            communication_rounds :
                the cumulative number of worker uploads by each iteration
            worker_upload_fraction :
                the average fraction of workers who upload each iteration
    """
    if rng is None:
        rng = np.random.default_rng()

    # Initialize the server
    server = fedavg_server(model_constructor(hparams), hparams.K)

    # Initialize the workers
    workers = [fedavg_worker(model_constructor(hparams)) for i in range(hparams.K)]

    # Partition the dataset into K splits and assign to worker

    # Note: In the real world we would not have access to the dataset as it would be distributed
    #      across all the worker devices. Here in simulation, we have access to the complete dataset
    #      and define the splits that go to each worker.

    X_train_splits,y_train_splits,split_weights = utils.split_training_data(X_train, y_train,  hparams.K, hparams.iid,
                                                                              rng)




    for i, worker in enumerate(workers):

        # worker.X_train = np.array(X_train[:, 3 * 3 * i:3 * 3 * (i + 1)])
        # worker.X_train = X_train
        #
        # worker.y_train = np.array(y_train[0:10000])
        # worker.X_test = np.array(X_test[:,3*3*i:3*3*(i+1)])
        # worker.X_test = X_test
        #
        # worker.y_test = np.array(y_test[0:5000])
        #
        worker.X_train = X_train_splits[i]
        worker.y_train = y_train_splits[i]


        #
        # a = pd.DataFrame(worker.X_train)
        # a.to_csv('train1.csv', index=None)
        # data = pd.read_csv('train1.csv')
        # data = np.array(data).astype('float32')
        # train2 = data.reshape((2000, 45))
        # b = pd.DataFrame(train2)
        # b.to_csv('train2.csv',index=None)
        # train2 = pd.read_csv('train2.csv')
        # train2 = np.array(train2).astype('float32')
        # X_train = train2
        #
        #
        #
        # c = pd.DataFrame(worker.X_test)
        # c.to_csv('test1.csv', index=None)
        # data2 = pd.read_csv('test1.csv')
        # data2 = np.array(data2).astype('float32')
        # test2 = data2.reshape((1000, 45))
        # b = pd.DataFrame(test2)
        # b.to_csv('test2.csv',index=None)
        # test2 = pd.read_csv('test2.csv')
        # test2 = np.array(test2).astype('float32')
        # X_test= test2
        #

        #
        # worker.X_train_fridge = np.array(X_train[:,i])
        # worker.X_train_television = np.array(X_train[:,i+5])
        # worker.X_train_kettle = np.array(X_train[:,i+10])
        # worker.X_train_microwave = np.array(X_train[:,i+15])
        # worker.X_train_washer_dryer = np.array(X_train[:,i+20])
        # worker.X_train_HTPC = np.array(X_train[:,i+25])
        # worker.X_train_radio = np.array(X_train[:,i+30])
        # worker.X_train_computer = np.array(X_train[:,i+35])
        # worker.X_train_fan = np.array(X_train[:,i+40])

        # a = pd.DataFrame(worker.X_train_fridge)
        # a.to_csv('train_fridge1.csv', index=None)
        #
        # # train_fridge = pd.read_csv('9_sumdata的副本.csv', header=None)
        # # train_fridge = np.array(train_fridge).astype('float32')
        # # worker.X_train_fridge = train_fridge
        #
        # data = pd.read_csv('train_fridge1.csv')
        # data = np.array(data).astype('float32')
        # train_fridge2 = data.reshape((6000, 45))
        # b = pd.DataFrame(train_fridge2)
        # b.to_csv('train_fridge2.csv',index=None)
        #
        # train_fridge2 = pd.read_csv('train_fridge2.csv')
        # train_fridge2 = np.array(train_fridge2).astype('float32')
        # worker.X_train_fridge = train_fridge2
        #
        #


        # worker.X_train = np.array(X_train[:, 3 * 3 * i:3 * 3 * (i + 1)])
        # worker.X_train = X_train
        #
        # worker.y_train = np.array(y_train[0:10000])
        # worker.X_test = np.array(X_test[:,3*3*i:3*3*(i+1)])
        # worker.X_test = X_test
        #
        # worker.y_test = np.array(y_test[0:5000])


    # Execute the iterations of FederatedAveraging and keep track of the number of communication rounds
    log = {"loss": [], "accuracy": [], "val_loss": [], "val_accuracy": [],
           "iteration": [], "communication_rounds": [], "worker_upload_fraction": []}
    communication_rounds = 0

    # Do initial evaluatio n of the randomly initialized global model as a baseline
    #utils.evaluate_and_log(log, server.global_model, X_train, y_train, X_test, y_test, 0, communication_rounds, hparams.K)

    m = int(np.ceil(hparams.C * hparams.K))  # Number of workers to use per iteration 每个迭代使用的工人数量

    # Note: In the real world each worker would perform its update in parallel on a separate device.
    #      Here in simulation, we can perform worker updates sequentially on the same device.
    for t in range(hparams.MAX_T):
        global_weights = server.global_model.get_weights()
        # Randomly pick the workers to be used for this iteration
        # 随机选择要用于此迭代的工作者
        worker_indices = set(rng.integers(hparams.K, size=m))

        # Perform the local update on each randomly selected worker starting from the global weights
        for wk_i in worker_indices:
            worker = workers[wk_i]
            worker.model.set_weights(global_weights)
            worker.model.fit(worker.X_train,worker.y_train, batch_size=hparams.B, epochs=hparams.E)
            #worker.model.fit(worker.X_train, worker.y_train, batch_size=9, epochs=100, validation_data=(worker.X_test, worker.y_test))

            #worker.model.fit(X_train, y_train, batch_size=9, epochs=100, validation_data=(X_test, y_test))
            # Upload the worker weights to the server
            server.worker_weights[wk_i] = worker.model.get_weights()
        communication_rounds += m

        # Average all the worker weights to get the updated global weights
        for i in range(len(global_weights)):
            global_weights[i] = np.sum(
                [split_weights[wk_i] * (server.worker_weights[wk_i][i] if wk_i in worker_indices else global_weights[i])
                 for wk_i in range(hparams.K)],
                axis=0)
        server.global_model.set_weights(global_weights)

        # Evaluate the global model on the train and validation sets on the evaluation interval
        if (t + 1) % hparams.evaluation_interval == 0:
            utils.evaluate_and_log(log, server.global_model, X_train, y_train, X_test, y_test, t + 1,
                                   communication_rounds, hparams.K)

            # Stop training when we have reached the target validation accuracy
            if log["val_accuracy"][-1] >= hparams.target_val_accuracy:
                break

    return server.global_model, log
